from gevent import monkey; monkey.patch_all()
from Lazy import LazyDict, AsyncCallable
from Cloud import utils as CloudUtils
from Parallel import utils as ParaUtils
from Face import modify
import inspect


def setup(only_gen_source=False):
    frame, path, line = inspect.getouterframes(inspect.currentframe())[1][:3]
    tree = modify(path, line, only_gen_source)
    if only_gen_source:
        print tree
        exit(0)
    new_gloabls = LazyDict(frame.f_globals)
    new_gloabls.update({"_product" : ParaUtils.product,
                        "_product2" : ParaUtils.product2,
                        "_pmap" : ParaUtils.pmap,
                        "_export": ParaUtils.export,
                        "_cloud": CloudUtils.Cloud,
                        "_cached": CloudUtils.CachedData,
                        "_async": AsyncCallable})
    code = compile(tree, 'path', 'exec')
    exec code in new_gloabls
    exit(0)


from gevent import monkey;monkey.patch_all()
import gevent
import inspect

class LazyDict(dict):
    def __init__(self, *args):
        super(LazyDict, self).__init__(*args)
        self.var_path = {}
        self.result_cache = {}

    def __getitem__(self, key):
        _result = super(LazyDict, self).__getitem__(key)
        if isinstance(_result, AsyncResult):
            _result = _result._result
            self.__setitem__(key, _result)
        return _result

class AsyncResult(object):
    next_id = 0
    def __init__(self, greenlet, r_vars=[], w_vars=[]):
        self.greenlet = greenlet
        self.r_vars = r_vars
        self.w_vars = w_vars
        self.__class__.next_id += 1
        self.uid = self.__class__.next_id

    def __getattr__(self, name):
        if name == "_result":
            self.greenlet.join()
            self.__dict__["_result"] = self.greenlet.value
            return self.__dict__["_result"]
        else:
            return getattr(self._result, name)


# class AsyncCallable(object):
#     def __init__(self, f):
#         self.f = f

#     def __call__(self, *args, **kargs):
#         greenlet = gevent.spawn(self.f, *args, **kargs)
#         return AsyncResult(greenlet)

import types

def AsyncCallable(r_vars, w_vars):
    def callWithVars(f):
        def func(*args, **kargs):
            greenlet = gevent.spawn(f, *args, **kargs)
            return AsyncResult(greenlet, r_vars, w_vars)
        return func
    return callWithVars


from uuid import uuid1

class AsyncFunction(object):
    def __init__(self, f):
        self.uuid = "%s.%s" % (uuid1().hex, f.func_name)
        self.func = types.FunctionType(f.func_code, 
                                       f.func_globals, 
                                       f.func_name, 
                                       f.func.func_defaults, 
                                       f.func_closure) 



import tokenize
import token

class TokenModifier(object):
    def __init__(self, f):
        self.tokens = tokenize.generate_tokens(f.readline)
        self.offset = 0
        self.dedents_patch_loc = []

    def visit(self):
        modified = []
        for toknum, tokval, tokbegin, tokend, tokline in self.tokens:
            # print (token.tok_name[toknum], tokval)
            if toknum != tokenize.COMMENT:
                modified.append((toknum, tokval))
            else:
                tokval_1 = tokval.strip(" \t#")
                tokbegin = tokbegin[0] + self.offset, tokbegin[1]
                tokend = tokend[0] + self.offset, tokend[1]
                handler_name = "%s_handler" % tokval_1.split()[0].lower()
                handler = getattr(self, handler_name, None)
                if handler:
                    dedents_new = len(modified)
                    new_tokens = handler(toknum, tokval_1, tokbegin, tokend, tokline)
                    self.offset += sum([1 for x in new_tokens if x[0]==tokenize.NEWLINE])
                    modified.extend(new_tokens)
                    dedents_old = len(modified)+1
                    self.dedents_patch_loc.append((dedents_new, dedents_old))
                else:
                    modified.append((toknum, tokval))
        # for x,y in modified:
        #     print (token.tok_name[x], y)
        # print self.dedents_patch_loc
        # print modified
        for x,y in self.dedents_patch_loc:
            # print modified[x], token.tok_name[modified[y][0]]
            if modified[y][0] in [tokenize.INDENT, tokenize.DEDENT]:
                modified.insert(x, modified[y])
                del modified[y+1]
        # print modified
        return tokenize.untokenize(modified)

    def parallelize_handler(self, toknum, tokval, tokbegin, tokend, tokline):
        new_tokens = []
        offset = 1
        for arg in tokval.split("with"):
            if arg.strip().startswith("const") or arg.strip().startswith("cached"):
                args = [x.strip(" \t,") for x in arg.strip().split()]
                args = [x for x in args if x not in ["and", "const", "consts", "cached"]]
                for v in args:
                    new_tokens.extend([
                                    (tokenize.NAME, v),
                                    (tokenize.OP, "="),
                                    (tokenize.NAME, "_cached"),
                                    (tokenize.OP, "("),
                                    (tokenize.NAME, v),
                                    (tokenize.OP, ")"),
                                    (tokenize.NEWLINE, "\n"),
                                    ])
                offset += len(args)
        if not hasattr(self, "parallel_token"):
            self.parallel_token = set()
        self.parallel_token.add(tokbegin[0]+offset)
        return new_tokens

    def remote_handler(self, toknum, tokval, tokbegin, tokend, tokline):
        return [
               (tokenize.OP, '@'),
               (tokenize.NAME, "_cloud"),
               (tokenize.NEWLINE, "\n"),
               (tokenize.OP, '@'),
               (tokenize.NAME, "_cached"),
               ]

    def async_handler(self, toknum, tokval, tokbegin, tokend, tokline):
        return [
               (tokenize.OP, '@'),
               (tokenize.NAME, '_async'),
               (tokenize.OP, '('),
               (tokenize.OP, '['),
               (tokenize.OP, ']'),
               (tokenize.OP, ','),
               (tokenize.OP, '['),
               (tokenize.OP, ']'),
               (tokenize.OP, ')'),
               ]




import os


CONF_LOC = ["/etc/briareus.conf",
            os.path.join(os.path.expanduser("~"), ".briareus.conf"),
            "./briareus.conf"]

class Config(object):
    def __init__(self):
        self.configs = {"port": 6379,
                        "queue_name": "Runtime",
                        "num_greenlet": 100}
                        
        for f in CONF_LOC:
            if os.path.exists(f):
                self.parse(f)

    def parse(self, path):
        with open(path, "r") as f:
            self.configs.update(eval(f.read()))

    def __getattr__(self, name):
        return self.configs[name]

config = Config()


if __name__ == '__main__':
    print config.host


from gevent import monkey; monkey.patch_all()
from Corellia.client import Client
from Corellia.kvstore import KVStore
from ..config import config
from uuid import uuid1
import Husky
from gevent import sleep

class Cloud(object):
    def __init__(self, f):
        self.f = f
        self.put_client = Client(config.host, "Briareus", pickler=Husky, serialize=True, interval=0.01)
        self.get_client = Client(config.host, "Briareus", pickler=Husky, serialize=False)
        # self.client = Client(config.host, "Briareus", pickler=Husky, serialize=True, interval=0.01)

    def __call__(self, *args):
        key = self.put_client.put_task("eval", (self.f, args))
        return self.get_client.get_result(key, block=True)
        # return Client(config.host, "Briareus", pickler=Husky, serialize=False).eval(self.f, args)
        # return self.client.eval(self.f, args)

    def __getstate__(self):
        return Husky.dumps(self.f)

    def __setstate__(self, state):
        self.f = Husky.loads(state) 

kvstore = KVStore(config.host, "CachedData")

class CachedData(object):
    def __init__(self, var):
        self.value = var
        self.put(self.value)

    def __getstate__(self):
        return self.id

    def __setstate__(self, state):
        self.id = state

    def __getattr__(self, name):
        if name == "value":
            self.__dict__["value"] = self.get(self.id)
            return self.__dict__["value"]
        else:
            return getattr(self.value, name)

    def put(self, value):
        self.id = uuid1().hex
        value = Husky.dumps(value)
        kvstore.set(self.id, value, serialize=True)

    def get(self, key):
        res = kvstore.get(key, serialize=False)
        while not res:
            sleep(0.01)
            res = kvstore.get(key, serialize=False)
        return Husky.loads(res)

    def __del__(self):
        pass


from gevent import monkey; monkey.patch_all()
from gevent.pool import Pool
import gevent
import inspect
from ..config import config


pool = Pool(config.num_greenlet)


def import_globals():
    frame, outerframe = [x[0] for x in inspect.getouterframes(inspect.currentframe())[1:3]]
    frame.f_globals.update(outerframe.f_globals)

def pmap(f, l):
    lets = []
    for i in l:
        lets.append(gevent.spawn(f, i))
    gevent.joinall(lets)
    return [i.value for i in lets]
    # return pool.map(f, l)

def export(ls):
    frame = inspect.getouterframes(inspect.currentframe())[1][0]
    for item in ls:
        frame.f_globals.update(item)

def product(*args):
    if len(args) == 1:
        return args[0]
    result = [[]]
    for arg in args:
        result = [x+[y] for x in result for y in arg]
    return result

def product2(iters, filters):
    result = [[]]
    for item, f in zip(iters, filters):
        if f:
            result = [x+[y] for x in result for y in item if f(x+[y])]
        else:
            result = [x+[y] for x in result for y in item]
    return result


if __name__ == '__main__':
    print product(range(10))






IN = 0
OUT = 1


class RoutineDependence(object):
    def __init__(self):
        self.varOperators = {}
        self.routineDeps = {}

    def add_routine(self, routine, ins, outs):
        r_deps = set()
        for v in ins:
            if v not in self.varOperators:
                self.varOperators[v] = [(IN, routine)]
            else:
                head = self.varOperators[v][-1]
                if isinstance(head, set):
                    for t, r in head:
                        r_deps.update(self.routineDeps[r])
                    head.add((IN, routine))
                elif head[0] == OUT:
                    r_deps.add(head[1])
                    self.varOperators[v].append((IN, routine))
                else:
                    r_deps.update(self.routineDeps[head[1]])
                    self.varOperators[v][-1] = set([(IN, routine), head])
        for v in outs:
            if v not in self.varOperators:
                self.varOperators[v] = [(OUT, routine)]
            else:
                head = self.varOperators[v][-1]
                if isinstance(head, set):
                    for t, r in head:
                        r_deps.add(r)
                else:
                    r_deps.add(head[1])
                self.varOperators[v].append((OUT, routine))
        self.routineDeps[routine] = r_deps

    def query_deps_by_routine(self, routine):
        return self.routineDeps[routine]

    def query_deps_by_vars(self, v, t):
        if v not in self.varOperators:
            return []
        head = self.varOperators[v][-1]
        if t == IN:
            if isinstance(head, set) or head[0] == IN:
                if len(self.varOperators[v]) > 1:
                    return [self.varOperators[v][-2][1]]
            else:
                return [head[1]]
        else:
            if isinstance(head, set):
                return [r0 for _, r0 in head]
            else:
                return [head[1]]
        return r

    def query_deps(self, routine, var, t):
        pass


if __name__ == '__main__':
    r = RoutineDependence()
    r.add_routine(1, [], ["a", "b"])
    r.add_routine(2, [], ["c", "d"])
    r.add_routine(3, ["a"], ["e", "f"])
    r.add_routine(4, ["a", "c"], ["f"])
    print r.routineDeps
    print r.varOperators
    print r.query_deps_by_routine(4)
    print r.query_deps_by_vars("f", OUT)







import cStringIO as StringIO
from token_mod import TokenModifier
from ast_mod import ASTModifier
import ast


import_str = """
from gevent import monkey; monkey.patch_all()
from Briareus.Parallel.utils import product as _product
from Briareus.Parallel.utils import product2 as _product2
from Briareus.Parallel.utils import pmap as _pmap
from Briareus.Parallel.utils import export as _export
from Briareus.Cloud.utils import Cloud as _cloud
from Briareus.Cloud.utils import CachedData as _cached
from Briareus.Lazy import AsyncCallable as _async
"""


def modify(file_path, pass_lines=0, source=None):
    with open(file_path, "r") as f:
        for x in xrange(pass_lines):
            f.readline()
        tokenMod = TokenModifier(f)
        modified = tokenMod.visit()
        parallel_tokens = getattr(tokenMod, "parallel_token", None)
    tree = ast.parse(modified)
    if parallel_tokens:
        tree = ast.fix_missing_locations(ASTModifier(parallel_tokens).visit(tree))
    if source:
        string = StringIO.StringIO()
        string.write(import_str)
        unparse.Unparser(tree, string)
        return string.getvalue()
    else:
        return tree


import ast
import unparse


class NamesMapper(ast.NodeTransformer):
    def __init__(self, maps):
        super(NamesMapper, self).__init__()
        self.maps = maps

    def visit_Name(self, node):
        if node.id in self.maps:
            return self.maps[node.id]
        else:
            return node

class ASTModifier(ast.NodeTransformer):
    def __init__(self, tokens):
        super(ASTModifier, self).__init__()
        self.var_count = 0
        self.tokens = tokens

    def make_name(self, prefix):
        name = "%s%d" % (prefix, self.var_count)
        self.var_count += 1
        return name

    def map_args(self, arg_items, body):
        if len(arg_items) == 1:
            arg = ast.arguments(args=[arg_items[0]], vararg=None, kwarg=None, defaults=[])
            return arg, body
        arg = ast.Name(id=self.make_name("_a"), ctx=ast.Load())
        maps = self.get_slice(arg, arg_items)
        if isinstance(body, list):
            body = [NamesMapper(maps).visit(i) for i in body]
        else:
            body = NamesMapper(maps).visit(body)
        arg = ast.arguments(args=[arg], vararg=None, kwarg=None, defaults=[])
        return arg, body

    def func_var_names(self, ast_node):
        node = ast.fix_missing_locations(ast_node)
        funcion_name = ast_node.name
        module = ast.Module(body=[ast_node])
        ctx = {}
        code = compile(module, '<string>', 'exec')
        eval(code, {}, ctx)
        function = ctx[funcion_name]
        return function.func_code.co_varnames
        

    def func_def(self, arg_items, body):
        args, body = self.map_args(arg_items, body)
        # body.append(ast.Return(value=ast.Call(func=ast.Name(id='locals', ctx=ast.Load()), args=[], keywords=[], starargs=None, kwargs=None)))
        name = self.make_name("_f")
        func = ast.FunctionDef(name=name, args=args, body=body, decorator_list=[])
        var_names = self.func_var_names(func)
        args_names = [arg.id for arg in args.args]
        global_names = [x for x in var_names if x not in args_names]
        if global_names:
            body = [ast.Global(names=global_names)] + body
        func = ast.FunctionDef(name=name, args=args, body=body, decorator_list=[])
        return name, func

    def map_func(self, func, iterable):
        return ast.Call(func=ast.Name(id='_pmap', ctx=ast.Load()), args=[func]+iterable, keywords=[], starargs=None, kwargs=None)

    def map_and_export(self, func, iterable):
        map_call = self.map_func(func, iterable)
        return ast.Call(func=ast.Name(id='_export', ctx=ast.Load()), args=[map_call], keywords=[], starargs=None, kwargs=None)

    def find_For_deepth(self, for_tree):
        yield for_tree.target, for_tree.iter
        if len(for_tree.body) == 1 and isinstance(for_tree.body[0], ast.For):
            for x in self.find_For_deepth(for_tree.body[0]):
                yield x
        else:
            yield for_tree.body

    def visit_For(self, for_tree):
        if for_tree.lineno not in self.tokens:
            return for_tree
        for node in ast.walk(for_tree):
            if isinstance(node, ast.Yield) or isinstance(node, ast.Return):
                return for_tree
        iters = [x for x in self.find_For_deepth(for_tree)]
        body = iters[-1]
        iters = iters[:-1]
        iterable = [self.product_func([x[1] for x in iters])]
        f_name, f_tree = self.func_def([x[0] for x in iters], body)
        func = ast.Name(id=f_name, ctx=ast.Load())
        map_tree = ast.Expr(value=self.map_func(func, iterable))
        return f_tree, map_tree

    def get_slice(self, src, args):
        maps = {}
        for i in xrange(len(args)):
            if isinstance(args[i], ast.Tuple):
                src_1 = ast.Subscript(value=src, slice=ast.Index(value=ast.Num(n=i)), ctx=ast.Load())
                maps.update(self.get_slice(src_1, args[i].elts))
            elif isinstance(args[i], ast.Name):
                maps[args[i].id] = ast.Subscript(value=src, slice=ast.Index(value=ast.Num(n=i)), ctx=ast.Load())
        return maps

    def visit_ListComp(self, node):
        if node.lineno not in self.tokens:
            return node
        arg_items = [x.target for x in node.generators]
        args, node = self.map_args(arg_items, node)
        iters = [x.iter for x in node.generators]
        filters = []
        for item in node.generators:
            if item.ifs:
                filters.append(ast.Lambda(args=args, body=item.ifs))
            else:
                filters.append(ast.Name(id='None', ctx=ast.Load()))
        if len(iters) == 1:
            if getattr(filters[0], "id", None) == "None":
                iterable = iters[0]
            else:
                iterable = ast.Call(func=ast.Name(id='filter', ctx=ast.Load()), args=[filters[0], iters[0]], keywords=[], starargs=None, kwargs=None)
        else:
            iterable = self.listcomp_func(iters, filters)
        return self.map_func(ast.Lambda(args=args, body=node.elt), [iterable])


    def product2_func(self, iters, filters):
        iters = ast.List(elts=iters, ctx=ast.Load())
        filters = ast.List(elts=filters, ctx=ast.Load())
        return ast.Call(func=ast.Name(id='_product2', ctx=ast.Load()), args=[iters, filters], keywords=[], starargs=None, kwargs=None)

    def product_func(self, iters):
        return ast.Call(func=ast.Name(id='_product', ctx=ast.Load()), args=iters, keywords=[], starargs=None, kwargs=None)



#!/usr/bin/env python

from setuptools import setup


requires = ["gevent>=0.13.0", "Corellia>=0.2.4", "Husky>=0.1.0"]
console_scripts=['Briareus-Worker=Briareus.Cloud.worker:run']

setup(name='Briareus',
      version='0.6.2',
      description='',
      author='Zhu Zhaomeng',
      author_email='zhaomeng.zhu@gmail.com',
      packages=['Briareus', 'Briareus.Cloud', 'Briareus.Parallel', 'Briareus.Face', 'Briareus.Lazy'],
      install_requires=requires,
      url="https://github.com/Tefx/Briareus",
      entry_points=dict(console_scripts=console_scripts),
      classifiers=[
          "Development Status :: 3 - Alpha",
          "Environment :: Console",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
          "Topic :: Utilities",
      ]
      )



from gevent import monkey; monkey.patch_all()
from utils import CachedData
from Corellia.taskqueue import TaskQueue

class Worker(object):
    def __init__(self, addr, path, **kargs):
        self.tq = TaskQueue(addr, path, **kargs)

    def run(self, cls, *args, **kargs):
        ins = cls(*args, **kargs)
        while 1:
            task =  self.tq.GET_TASK()
            if not task:
                continue
            key, method, args = task
            func = getattr(ins, method, None)
            if func:
                # try:
                result = func(*args)
                # except Exception, e:
                #   result = str(e)
            else:
                result = "No Such Method!"
            self.tq.PUT_RESULT(key, result) 


class Runtime(object):
    def __init__(self):
        self.cache = {}

    def getvalue(self, v):
        if isinstance(v, CachedData):
            if v.id in self.cache:
                return self.cache[v.id]
            else:
                self.cache[v.id] = v.value
                return v.value
        else:
            return v

    def eval(self, f, args):
        f = self.getvalue(f)
        return f(*map(self.getvalue, args))

def run():
    from Corellia.worker import WorkerPool
    import Husky
    from ..config import config
    Worker(config.host, "Briareus", pickler=Husky, serialize=False).run(Runtime)


